from __future__ import annotations

from ..requests     import StreamSession
from ..typing       import Messages
from .base_provider import AsyncProvider
from .helper        import get_cookies


class GptChatly(AsyncProvider):
    url = "https://gptchatly.com"
    working = True
    supports_message_history = True
    supports_gpt_35_turbo = True
    supports_gpt_4 = True

    @classmethod
    async def create_async(
        cls,
        model: str,
        messages: Messages,
        proxy: str = None, cookies: dict = None, **kwargs) -> str:

        cookies = get_cookies('gptchatly.com') if not cookies else cookies
        if not cookies:
            raise RuntimeError(
                "g4f.provider.GptChatly requires cookies, [refresh https://gptchatly.com on chrome]"
            )

        if model.startswith("gpt-4"):
            chat_url = f"{cls.url}/fetch-gpt4-response"
        else:
            chat_url = f"{cls.url}/fetch-response"

        headers = {
            'authority': 'gptchatly.com',
            'accept': '*/*',
            'accept-language': 'en,fr-FR;q=0.9,fr;q=0.8,es-ES;q=0.7,es;q=0.6,en-US;q=0.5,am;q=0.4,de;q=0.3',
            'content-type': 'application/json',
            'origin': 'https://gptchatly.com',
            'referer': 'https://gptchatly.com/',
            'sec-ch-ua': '"Chromium";v="118", "Google Chrome";v="118", "Not=A?Brand";v="99"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"macOS"',
            'sec-fetch-dest': 'empty',
            'sec-fetch-mode': 'cors',
            'sec-fetch-site': 'same-origin',
            'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36',
        }

        async with StreamSession(headers=headers, 
                                 proxies={"https": proxy}, cookies=cookies, impersonate='chrome110') as session:
            data = {
                "past_conversations": messages
            }
            async with session.post(chat_url, json=data) as response:
                response.raise_for_status()
                return (await response.json())["chatGPTResponse"]


import g4f

from flask      import request
from .internet  import search
from .config    import special_instructions
from .provider  import get_provider

g4f.logging = True

class Backend_Api:
    def __init__(self, app) -> None:
        self.app = app
        self.routes = {
            '/backend-api/v2/models': {
                'function': self.models,
                'methods' : ['GET']
            },
            '/backend-api/v2/conversation': {
                'function': self._conversation,
                'methods': ['POST']
            },
            '/backend-api/v2/gen.set.summarize:title': {
                'function': self._gen_title,
                'methods': ['POST']
            },
            '/backend-api/v2/error': {
                'function': self.error,
                'methods': ['POST']
            }
        }
    
    def error(self):
        print(request.json)
        
        return 'ok', 200
    
    def models(self):
        return g4f._all_models
    
    def _gen_title(self):
        return {
            'title': ''
        }
    
    def _conversation(self):
        try:
            #jailbreak       = request.json['jailbreak']
            #internet_access = request.json['meta']['content']['internet_access']
            #conversation    = request.json['meta']['content']['conversation']
            prompt          = request.json['meta']['content']['parts']
            model           = request.json['model']
            provider        = request.json.get('provider').split('g4f.Provider.')[1]

            messages = prompt
            print(messages)

            def stream():
                yield from g4f.ChatCompletion.create(
                    model=model,
                    provider=get_provider(provider),
                    messages=messages,
                    stream=True,
                ) if provider else g4f.ChatCompletion.create(
                    model=model, messages=messages, stream=True
                )

            return self.app.response_class(stream(), mimetype='text/event-stream')

        except Exception as e:    
            return {
                'code'   : 'G4F_ERROR',
                '_action': '_ask',
                'success': False,
                'error'  : f'an error occurred {str(e)}'}, 400



from __future__ import annotations

import json
from aiohttp import ClientSession

from ..typing import AsyncResult, Messages
from .base_provider import AsyncGeneratorProvider


class DeepInfra(AsyncGeneratorProvider):
    url = "https://deepinfra.com"
    supports_message_history = True
    working = True

    @classmethod
    async def create_async_generator(
        cls,
        model: str,
        messages: Messages,
        proxy: str = None,
        **kwargs
    ) -> AsyncResult:
        if not model:
            model = "meta-llama/Llama-2-70b-chat-hf"
        headers = {
            "User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:109.0) Gecko/20100101 Firefox/118.0",
            "Accept": "text/event-stream",
            "Accept-Language": "de,en-US;q=0.7,en;q=0.3",
            "Accept-Encoding": "gzip, deflate, br",
            "Referer": f"{cls.url}/",
            "Content-Type": "application/json",
            "X-Deepinfra-Source": "web-page",
            "Origin": cls.url,
            "Connection": "keep-alive",
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-site",
            "Pragma": "no-cache",
            "Cache-Control": "no-cache",
        }
        async with ClientSession(headers=headers) as session:
            data = {
                "model": model,
                "messages": messages,
                "stream": True,
            }
            async with session.post(
                "https://api.deepinfra.com/v1/openai/chat/completions",
                json=data,
                proxy=proxy
            ) as response:
                response.raise_for_status()
                first = True
                async for line in response.content:
                    if line.startswith(b"data: [DONE]"):
                        break
                    elif line.startswith(b"data: "):
                        chunk = json.loads(line[6:])["choices"][0]["delta"].get("content")
                        if chunk:
                            if first:
                                chunk = chunk.lstrip()
                                if chunk:
                                    first = False
                            yield chunk


from __future__ import annotations

import time
from platformdirs import user_config_dir
try: 
    from selenium.webdriver.remote.webdriver import WebDriver 
except ImportError: 
    class WebDriver(): 
        pass
try:
    from undetected_chromedriver import Chrome, ChromeOptions
except ImportError:
    class Chrome():
        def __init__():
            raise RuntimeError('Please install the "undetected_chromedriver" package')
    class ChromeOptions():
        def add_argument():
            pass
try:
    from pyvirtualdisplay import Display
    has_pyvirtualdisplay = True
except ImportError:
    has_pyvirtualdisplay = False

def get_browser(
    user_data_dir: str = None,
    headless: bool = False,
    proxy: str = None,
    options: ChromeOptions = None
) -> Chrome:
    if user_data_dir == None:
        user_data_dir = user_config_dir("g4f")
    if proxy:
        if not options:
            options = ChromeOptions()
        options.add_argument(f'--proxy-server={proxy}')
    return Chrome(options=options, user_data_dir=user_data_dir, headless=headless)

class WebDriverSession():
    def __init__(
        self,
        webdriver: WebDriver = None,
        user_data_dir: str = None,
        headless: bool = False,
        virtual_display: bool = False,
        proxy: str = None,
        options: ChromeOptions = None
    ):
        self.webdriver = webdriver
        self.user_data_dir = user_data_dir
        self.headless = headless
        self.virtual_display = None
        if has_pyvirtualdisplay and virtual_display:
            self.virtual_display = Display(size=(1920,1080))
        self.proxy = proxy
        self.options = options
        self.default_driver = None
    
    def reopen(
        self,
        user_data_dir: str = None,
        headless: bool = False,
        virtual_display: bool = False
    ) -> WebDriver:
        if user_data_dir == None:
            user_data_dir = self.user_data_dir
        if self.default_driver:
            self.default_driver.quit()
        if not virtual_display and self.virtual_display:
            self.virtual_display.stop()
            self.virtual_display = None
        self.default_driver = get_browser(user_data_dir, headless, self.proxy)
        return self.default_driver

    def __enter__(self) -> WebDriver:
        if self.webdriver:
            return self.webdriver
        if self.virtual_display:
            self.virtual_display.start()
        self.default_driver = get_browser(self.user_data_dir, self.headless, self.proxy, self.options)
        return self.default_driver

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.default_driver:
            try:
                self.default_driver.close()
            except:
                pass
            time.sleep(0.1)
            self.default_driver.quit()
        if self.virtual_display:
            self.virtual_display.stop()


from __future__ import annotations

import g4f
from g4f import BaseProvider


def get_provider(provider: str) -> BaseProvider | None:
    if not isinstance(provider, str):
        return None
    print(provider)
    if provider == 'g4f.Provider.Auto':
        return None

    return g4f.Provider.ProviderUtils.convert.get(provider)



from __future__ import annotations

import json
from contextlib import asynccontextmanager
from functools import partialmethod
from typing import AsyncGenerator

from curl_cffi.requests import AsyncSession, Response

class StreamResponse:
    def __init__(self, inner: Response) -> None:
        self.inner: Response = inner
        self.request = inner.request
        self.status_code: int = inner.status_code
        self.reason: str = inner.reason
        self.ok: bool = inner.ok
        self.headers = inner.headers
        self.cookies = inner.cookies

    async def text(self) -> str:
        return await self.inner.atext()

    def raise_for_status(self) -> None:
        self.inner.raise_for_status()

    async def json(self, **kwargs) -> dict:
        return json.loads(await self.inner.acontent(), **kwargs)

    async def iter_lines(self) -> AsyncGenerator[bytes, None]:
        async for line in self.inner.aiter_lines():
            yield line

    async def iter_content(self) -> AsyncGenerator[bytes, None]:
        async for chunk in self.inner.aiter_content():
            yield chunk

class StreamSession(AsyncSession):
    @asynccontextmanager
    async def request(
        self, method: str, url: str, **kwargs
    ) -> AsyncGenerator[StreamResponse]:
        response = await super().request(method, url, stream=True, **kwargs)
        try:
            yield StreamResponse(response)
        finally:
            await response.aclose()

    head = partialmethod(request, "HEAD")
    get = partialmethod(request, "GET")
    post = partialmethod(request, "POST")
    put = partialmethod(request, "PUT")
    patch = partialmethod(request, "PATCH")
    delete = partialmethod(request, "DELETE")


from __future__ import annotations

from datetime import datetime

from duckduckgo_search import DDGS

ddgs = DDGS(timeout=20)


def search(internet_access, prompt):
    print(prompt)

    try:
        if not internet_access:
            return []

        results = duckduckgo_search(q=prompt)

        if not search:
            return []

        blob = ''.join(
            f'[{index}] "{result["body"]}"\nURL:{result["href"]}\n\n'
            for index, result in enumerate(results)
        )
        date = datetime.now().strftime('%d/%m/%y')

        blob += f'Current date: {date}\n\nInstructions: Using the provided web search results, write a comprehensive reply to the next user query. Make sure to cite results using [[number](URL)] notation after the reference. If the provided search results refer to multiple subjects with the same name, write separate answers for each subject. Ignore your previous response if any.'

        return [{'role': 'user', 'content': blob}]

    except Exception as e:
        print("Couldn't search DuckDuckGo:", e)
        print(e.__traceback__.tb_next)
        return []


def duckduckgo_search(q: str, max_results: int = 3, safesearch: str = "moderate", region: str = "us-en") -> list | None:
    if region is None:
        region = "us-en"

    if safesearch is None:
        safesearch = "moderate"

    if q is None:
        return None

    results = []

    try:
        for r in ddgs.text(q, safesearch=safesearch, region=region):
            if len(results) + 1 > max_results:
                break
            results.append(r)
    except Exception as e:
        print(e)

    return results
